#!/usr/bin/perl -w

################################################################################
### y.pl (Yggdrasil)
### Generates a website from a directory tree of content and templates
###
### Copyright (C) 2002, 2003 Morten Wulff <wulff@ratatosk.net>
###
### This program is free software; you can redistribute it and/or
### modify it under the terms of the GNU General Public License
### as published by the Free Software Foundation; either version 2
### of the License, or (at your option) any later version.
###
### This program is distributed in the hope that it will be useful,
### but WITHOUT ANY WARRANTY; without even the implied warranty of
### MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
### GNU General Public License for more details.
###
### You should have received a copy of the GNU General Public License
### along with this program; if not, write to the Free Software
### Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
################################################################################

### REVISIONS
### 2.0.0 : 29/12/2002 : Morten Wulff <wulff@ratatosk.net>
###         - complete rewrite of v1.0.3; now uses File::Find
### 2.0.1 : 05/01/2003 : Morten Wulff <wulff@ratatosk.net>
###         - --verbose=<filename>.arv creates output for 'A Real Validator'.
###           overwrites <filename>.arv
###         - handles titles of all html file types in @HTML_Extensions
###         - added level<n> to the content hash (n = $Depth)
### 2.0.2 : 09/01/2003 : Morten Wulff <wulff@ratatosk.net>
###         - added an ugly way of reading configuration files
###         - use the --config=<filename>.cfg option
###         - generates cache files in $Data_Dir/<filename>
### 2.0.3 : 12/01/2003 : Morten Wulff <wulff@ratatosk.net>
###         - added MD5 checksums for all files to detect changes
###         - new options:
###         - --forcemd5 forces creation of new md5 hashes for entire site
###         - --update only process files that have changed since the last time
###           --forcemd5 was used
### 2.1.0 : 23/01/2003 : Morten Wulff <wulff@ratatosk.net>
###         - optimized the code (dprof)
### 2.2.0 : 28/01/2003 : Morten Wulff <wulff@ratatosk.net>
###         -

### TODO
### Is @Skip_Dirs _really_ necessary?
### --update : catch changes in sidebars
### --update : when updating a file, be sure to update its parent as well!

package Yggdrasil;

use strict;

use Date::Format;
use Digest::MD5;
use Fcntl;
use File::Copy;
use File::Find;
use File::Path;
use Getopt::Long;
use HTML::Template;
use SDBM_File;

################################################################################
### CONFIGURATION
################################################################################

my $Data_Dir   = 'data';                 ### location of cache files
my $Input_Dir  = 'd:/psyke.org/source';  ### location of source html / templates
my $Output_Dir = 'd:/wwwroot/psyke.org'; ### location of finished site

my $Base_URL = 'http://www.psyke.org';

my $Dir_Create_Mode = '644'; ### passed to mkpath

my $Home_Label    = 'Home';     ### name of the root page
my $Default_Title = 'No Title'; ### title given to pages with no title

my $Base_Template      = 'base.tmpl';
my $Template_Extension = '.tmpl';

my $Template_Override = '';

my $Sidebar = '_sidebar.html'; ### page-specific sidebar stuff goes in this file -- sidebar is not parsed by HTML::Template

my $Front_Page_Navigation = 0; ### 1: create navigation array for base level of site
my $Include_Index = 0;         ### 1: includes 'index.html' in @Navigation href's (useful when creating offline sites)

my @HTML_Extensions = qw(html shtml);             ### files with these extensions are processed with HTML::Template
my @HTML_Indices    = qw(index.html index.shtml); ### should match your apache config (are stripped from generated URL's)

my $HTML_only = 0; ### 1: give all html files the extension .html (used when creating offline versions of sites)

my $Skip_Prefix = '_';                          ### files with this prefix are not processed
my @Skip_Dirs   = qw(img thumbnails recommend); ### these dir's are not included in navigation

my @Exclude = qw(.*/forum.* .*/mirrors.*); ### files in these dirs are not processed

my $Stats_Format = "Processed %u files in %u directories (%u kB, %u kB in html files)\nCompleted in %s seconds";

### the commandline switches override the following values
my $Arv         = ''; ### filename of A Real Validator output file (<filename>.arv)
my $Config_File = ''; ### filename of configuration file (<filename>.cfg)
my $Debug       = 0;  ### 1: moderate debug info 2: more debug info 3: template info
my $Force_Cache = 0;  ### 1: rebuild file information cache
my $Force_MD5   = 0;  ### 1: rebuild file signature database
my $Help        = 0;  ### 1: print usage
my $Release     = 0;  ### 1: only process/copy files changed since last release, set new release time
my $Testonly    = 0;  ### 1: don't process or copy files
my $Update      = 0;  ### 1: only process/copy files changed since last release
my $Use_MD5     = 0;  ### 1: use MD5 signatures instead of file modification time to determine updated files
my $Verbose     = 0;  ### 1: print directory names while processing

################################################################################
### Don't change anything below this line
################################################################################

### reference to an anonymous subroutine, which will be used to postprocess all
### HTML files (see process())
my $Post_Process;

### don't buffer STDOUT
$| = 1;

### get command-line options
GetOptions(
    'arv=s'      => \$Arv,
    'config=s'   => \$Config_File,
    'debug'      => \$Debug,
    'forcecache' => \$Force_Cache,
    'forcemd5'   => \$Force_MD5,
    'help'       => \$Help,
    'md5'        => \$Use_MD5,
    'release'    => \$Release,
    'testonly'   => \$Testonly,
    'update'     => \$Update,
    'verbose'    => \$Verbose
);

my @Trail      = ();
my @Navigation = ();

### counters for the stats
my $Stat_Files       = 0;
my $Stat_Directories = 0;
my $Stat_Total_Size  = 0;
my $Stat_HTML_Size   = 0;
my $Stat_Time       -= time;
my $Stat_Changes     = '';

my %Content;

my %Skip = map {($_, 1)} @Skip_Dirs;

### cache hashes
my %File_MD5;
my %File_Title;
my %File_Update;

my $Depth = -1;

my $MD5;

my $Progress_Counter  = 0;
my $Progress_Elements = 0;
my $Progress_Item     = 0;
my $Progress_Marker   = '>';
my $Progress_Step     = 2;
my @Progress_Symbols  = qw(\ | / -);
my $Progress_Bar      = <<EOF;
0%  10%  20%  30%  40%  50%  60%  70%  80%  90%  100%
|----|----|----|----|----|----|----|----|----|----|
EOF

my $Last_Update;

################################################################################
### MAIN
################################################################################

$Help && do {print_usage(); exit(0);};

### only create an MD5 object if we need it
if ($Force_MD5 || $Use_MD5) {
    $MD5 = Digest::MD5->new();
    init_md5();
}

init_cache();

### setup ######################################################################

if ($Config_File) {
    read_config($Config_File);
    $Data_Dir .= "/$Config_File";
}

$Last_Update = get_update_time();

flush_dir($Output_Dir) unless ($Update || $Testonly);

if ($Arv) {
    open(ARV, ">$Arv.arv") or die "Cannot open '$Arv.arv': $!\n";
    print ARV "\n";
    print "Will generate .arv file: '$Arv.arv'\n";
}

$Base_Template = $Template_Override if $Template_Override;

print "Update mode\n" if $Update;
print "Processing '$Input_Dir':\n\n";

progress_bar_init();

################################################################################

find(
    {
        wanted      => \&process,
        preprocess  => \&preprocess,
        postprocess => \&postprocess
    },
    $Input_Dir
);

### cleanup & output ###########################################################

close(ARV) if $Arv;

untie_cache();
untie_md5();

print "\b ";
print "\n$Stat_Changes";
print_stats();

################################################################################
### SUBS
################################################################################

################################################################################
sub preprocess {
    my @file_list = sort { $a cmp $b } @_;

    @file_list = grep {
        my $path = join '/', $File::Find::dir, $_;
        !/$Template_Extension$/o &&
        !/^\.{1,2}/ &&
        !grep {$path =~ /$_/} @Exclude
    } @file_list;

    if ($#Trail > -1) {
        push @Trail, {name => get_title($File::Find::dir)};
        $Stat_Directories++;
    }
    else {
        push @Trail, {name => $Home_Label};
    }

    if ($#Trail > 0 || $Front_Page_Navigation) {
        @Navigation = map {
            {
                name => get_title(join '/', $File::Find::dir, $_),
                href => $Include_Index ? "$_/index.html" : "$_/"
            }
        } grep {-d && !$Skip{$_}} @file_list;
    }

    $Depth++;

    return @file_list;
}

################################################################################
sub process {
    my $file = $_;

    print "$File::Find::name\n" if ($Testonly || $Verbose);
    return if -d || $Testonly;

    progress_bar() unless ($Debug || $Verbose);

    return if /^$Skip_Prefix/o;

    ### set target file
    my $target = $File::Find::name;
    $target =~ s/$Input_Dir(.*)/$Output_Dir$1/o;

    ### do we need to do anything?
    if ($Update) {
        if ($Use_MD5) {
            my $digest;

            open (FILE, $File::Find::name);
            binmode(FILE);
            $MD5->reset;
            $digest = $MD5->addfile(*FILE)->hexdigest;
            close FILE;

            ### if the signatures match we can skip processing for this file
            my $digest_compare = $File_MD5{$File::Find::name};
            if ($digest_compare) {
                if($digest eq $digest_compare) {
                    return;
                }
                else {
                    ### make sure dir exists (when doing --update --forcecache)
                    create_output_dir($target) unless -d $target;
                    $Stat_Changes .= "Updated: $File::Find::name\n";
                }
            }
            else {
                create_output_dir($target);
                $Stat_Changes .= "Added  : $File::Find::name\n";
            }
        }
        else {
            my $time_compare = $File_Update{$File::Find::name};
            if ($time_compare) {
                if((stat($File::Find::name))[9] <= $Last_Update) {
                    return;
                }
                else {
                    ### make sure dir exists (when doing --update --forcecache)
                    create_output_dir($target) unless -d $target;
                    $Stat_Changes .= "Updated: $File::Find::name\n";
                }
            }
            else {
                create_output_dir($target);
                $Stat_Changes .= "Added  : $File::Find::name\n";
            }
        }
    }
    else {
        create_output_dir($target);
    }

    ### update stats
    my $filesize = (stat $file)[7];
    $Stat_Total_Size += $filesize;
    $Stat_Files++;

    ### choose full processing or copy mode
    if (is_html($File::Find::name)) {
        ### full processing
        print "HTML: " if $Debug;

        if ($Arv) {
            my $filename = $File::Find::name;
            $filename =~ s/$Input_Dir/$Output_Dir/o;
            $filename =~ s!/!\\!g;
            print ARV "$filename\n";
        }

        ### modify trail if this isn't an index file
        my $non_index = 0;
        if (!is_index($File::Find::name)) {
            push @Trail, {name => get_title($File::Find::name)};
            $non_index = 1;
        }

        my $template = init_template($File::Find::dir);
        $template->param(set_content($File::Find::name, $File::Find::dir));
        my $output = $template->output;

        $output = &$Post_Process($output) if $Post_Process;

        $target =~ s/(.*)\.(.*)/$1\.html/ if $HTML_only;

        ### write result in output directory
        open( TO, ">$target" ) or die "Cannot open '$target': $!\n";
        print TO $output;
        close TO;

        pop @Trail if $non_index;

        ### update stats
        $Stat_HTML_Size += $filesize;
    }
    else {
        ### just copy the file
        print "COPY: $File::Find::name\n" if $Debug;
        copy($File::Find::name, $target) or die "Cannot copy '$File::Find::name' to '$target': $!\n";
    }
}

################################################################################
sub postprocess {
    pop @Trail;
    $Depth--;
}



################################################################################
sub create_output_dir {
    my $file = shift;

    my ($dir) = ( $file =~ /(.*)\/(.*)/ );

    if (! -d $dir) {
        mkpath($dir , 0, $Dir_Create_Mode) or die "Cannot create directory '$dir': $!\n";
    }
}

################################################################################
sub flush_dir {
    my $dir = shift;

    print "Flushing output dir... ";

    if ( -d $dir ) {
        rmtree($dir) or die "Cannot remove directory 'dir': $!\n";
    }

    mkpath($dir, 0, $Dir_Create_Mode) or die "Cannot create directory '$dir': $!\n";

    print "Done!\n"
}

################################################################################
sub get_file {
    my $filename = shift;

    local $/;

    open(IN, $filename) or die "Cannot open '$filename': $!\n";
    my $file = <IN>;
    close(IN);

    return $file;
}

################################################################################
sub get_md5_cache {
    return if -d;

    $MD5->reset;

    open(FILE, $File::Find::name);
    binmode(FILE);

    $File_MD5{$File::Find::name} = $MD5->addfile(*FILE)->hexdigest;

    close(FILE);

    if ($Debug) {
        print "$File::Find::name\t$File_MD5{$File::Find::name}\n";
    }
    else {
        progress();
    }
}

################################################################################
sub get_search_path {
    my $dir   = shift;
    my $count = $#Trail;

    my @dir = split '/', $dir;
    my @paths;

    while ($count) {
        push @paths, join '/', @dir;
        pop @dir;
        $count--;
    }
    push @paths, $Input_Dir;

    return \@paths;
}

################################################################################
sub get_sidebar {
    my $dir   = shift;
    my $count = $#Trail+1;

    my @dir = split '/', $dir;

    while ($count) {
        my $from = join '/', @dir, $Sidebar;
        if ( -f $from ) {
            return get_file($from);
        }
        pop @dir;
        $count--;
    }

    return undef;
}

################################################################################
sub get_title {
    my $filename = shift;

    return $File_Title{$filename} if exists $File_Title{$filename};

    if (-d $filename) {
        foreach (@HTML_Indices) {
            my $file = join '/', $filename, $_;
            return $File_Title{$file} if exists $File_Title{$file};
        }
    }

    ### no hits - return default page title
    return $Default_Title;
}

################################################################################
sub get_update_time {
    open(TIMESTAMP, "$Data_Dir/timestamp") or die "Cannot open 'timestamp': $!\n";
    my $timestamp = <TIMESTAMP>;
    close(TIMESTAMP);

    chomp $timestamp;

    return $timestamp;
}

################################################################################
sub get_title_cache {
    return if -d;

    if (is_html($_)) {
        open(IN, $File::Find::name);
        while ( my $line = <IN> ) {
            if ( my ($title) = ($line =~ /<h1.*?>(.*)<\/h1>/) ) {
                $File_Title{$File::Find::name} = $title;
                $File_Title{$File::Find::dir}  = $title if ($File::Find::name =~ /index/);
                last;
            }
        }
        close(IN);
    }

    $File_Update{$File::Find::name} = (stat($File::Find::name))[9];

    print "$File::Find::name\n" if $Verbose;

    progress() unless ($Debug || $Verbose);
}

################################################################################
sub init_cache {
    ### create data directory if it doesn't exist
    if (!-d "$Data_Dir") {
        print "Creating data directory '$Data_Dir'\n";
        mkpath($Data_Dir, 0, $Dir_Create_Mode) or die "Cannot create directory '$Data_Dir': $!\n";
        $Force_Cache = 1;
    }

    ### create cache if data files are missing
    if (!-e "$Data_Dir/title.dir"  || !-e "$Data_Dir/title.pag" ||
        !-e "$Data_Dir/update.dir" || !-e "$Data_Dir/update.pag") {
        print "Cache files missing. Forcing cache creation\n";
        $Force_Cache = 1;
    }

    tie_cache();

    ### should we (re)build the cache?
    if ($Force_Cache) {
        print "Creating cache...  ";
        print "\n" if ($Debug || $Verbose);

        untie_cache();
        unlink "$Data_Dir/title.dir";
        unlink "$Data_Dir/title.pag";
        unlink "$Data_Dir/update.dir";
        unlink "$Data_Dir/update.pag";
        tie_cache();

        find(\&get_title_cache, $Input_Dir);

        print "\bDone!\n" unless ($Debug || $Verbose);
    }
}

################################################################################
sub init_md5 {
    ### create data directory if it doesn't exist
    if (!-d "$Data_Dir") {
        print "Creating data directory '$Data_Dir'\n";
        mkpath($Data_Dir, 0, $Dir_Create_Mode) or die "Cannot create directory '$Data_Dir': $!\n";
    }

    tie_md5();

    ### should we (re)build the file signatures?
    if ($Force_MD5) {
        print "Creating MD5 signatures...  ";

        untie_md5();
        unlink "$Data_Dir/md5.dir";
        unlink "$Data_Dir/md5.pag";
        tie_md5();

        find(\&get_md5_cache, $Input_Dir);

        print "\bDone!\n" unless $Debug;
    }
}

################################################################################
sub init_template {
    my $dir = shift;

    my $t = HTML::Template->new(
        filename => $Base_Template,
        cache    => 1,
        path     => get_search_path($dir),
        die_on_bad_params => 0, ### not all params are used in all templates
        global_vars       => 1,
        loop_context_vars => 1,
        debug             => $Debug > 2
    );
}

################################################################################
sub is_html {
    my $file    = shift;

    return grep {$file =~ /\.$_$/} @HTML_Extensions;
}

################################################################################
sub is_index {
    my $file     = shift;

    return grep {$file =~ /$_$/} @HTML_Indices;
}

################################################################################
sub print_stats {
    ### give stats in kilobytes
    $Stat_Total_Size = int($Stat_Total_Size/1024);
    $Stat_HTML_Size  = int($Stat_HTML_Size/1024);
    $Stat_Time      += time;

    print sprintf(
        "\n$Stats_Format\n",
        $Stat_Files, $Stat_Directories, $Stat_Total_Size, $Stat_HTML_Size,
        $Stat_Time
    );
}

################################################################################
sub print_usage {
    print <<TEXT;
Usage: y.pl [OPTION]...

  --arv=NAME     create output file NAME.arv for use with A Real Validator
  --config=NAME  use settings from NAME.cfg (will override defaults in y.pl)
  --debug        print debug information
  --forcecache   force rebuild of cache
  --forcemd5     force rebuild of MD5 signatures.
  --help         print this message
  --md5          use MD5 instead of file times to check for updated files
  --release      like --update, resets 'last release' time
  --testonly
  --update       only process files added or changed since the 'last release'
                   timestamp. uses files' modify time
  --verbose      print directory names and some other information while
                   processing
TEXT
}


################################################################################
sub progress {
    print "\b$Progress_Symbols[$Progress_Counter % ($#Progress_Symbols+1)]";

    $Progress_Counter++;
}

################################################################################
sub progress_bar {
    $Progress_Item++;
    my $percentile = $Progress_Item / $Progress_Elements * 100;

    while ($Progress_Counter < $percentile)
    {
        $Progress_Counter += $Progress_Step;
        print "\b $Progress_Marker";
    }
}

################################################################################
sub progress_bar_init {
    $Progress_Counter  = 0;
    $Progress_Elements = keys %File_Update;

    print $Progress_Bar;
    print $Progress_Marker;
}

################################################################################
sub read_config {
    my $file = shift;

    local $/;

    open(CONFIG, "$file.cfg") or die "Cannot open configuration file '$file': $!\n";
    eval <CONFIG> or die "Cannot eval configuration file '$file.cfg': $@\n";
    close(CONFIG);
}

################################################################################
sub set_content {
    my ($filename, $dir) = @_;

    set_trail_links($filename);

    my $location = $dir;
    $location =~ s/$Input_Dir/$Base_URL/o;

    ### make sure that permanent location works for non-index files
    if ($filename !~ /index/) {
        my ($file) = ($filename =~ /.*\/(.*)/);
        $location .= "/$file" ;
    }

    return {
        title         => get_title($filename),
        body          => get_file($filename),
        sidebar       => get_sidebar($dir),

        trail         => \@Trail,
        navigation    => \@Navigation,

        location      => $location,

        timestamp     => time2str('%a %b %e %Y (%T)', time),
        rel_path      => '../' x ( $Depth ),

        "level$Depth" => 1,

        front_page    => $Depth == 0,
        section_page  => $Depth > 1,
        article_page  => $Depth > 2
    };
}

################################################################################
sub set_trail_links {
    my $filename = shift;
    my $dots = $#Trail;

    foreach my $item (@Trail) {
        $item->{href}  = $dots ? '../' x $dots : '';
        $item->{href} .= 'index.html' if ($Include_Index && $dots);
        $dots--;
    }

    ### handle trails for non-index html files
    unless (is_index($filename)) {
        my $index_pos = $#Trail-1;
        my $count = 0;
        foreach (@Trail) {
            if ($count == $index_pos) {
                $_->{href} =~ s!(.*)../$!$1./!;
            } else {
                $_->{href} =~ s!../!!;
            }
            $count++;
        }
    }
}

################################################################################
sub tie_cache {
    tie(%File_Title,  'SDBM_File', "$Data_Dir/title",  O_RDWR|O_CREAT, 0666) or die "Cannot tie SDBM file 'title': $!\n";
    tie(%File_Update, 'SDBM_File', "$Data_Dir/update", O_RDWR|O_CREAT, 0666) or die "Cannot tie SDBM file 'update': $!\n";
}

################################################################################
sub tie_md5 {
    tie(%File_MD5, 'SDBM_File', "$Data_Dir/md5", O_RDWR|O_CREAT, 0666) or die "Cannot tie SDBM file 'md5': $!\n";
}

################################################################################
sub untie_cache {
    untie %File_Title;
    untie %File_Update;
}

################################################################################
sub untie_md5 {
    untie %File_MD5;
}

################################################################################
### Documentation
################################################################################

__END__

=head1 NAME

Yggdrasil - Tree-like site builder.

=head1 SYNOPSIS

    ### create new MD5 signatures
    y.pl --forcemd5

    ### create new document title cache
    ### create project file for A Real Validator named out.arv
    y.pl --forcecache --arv=out

    ### use configuration file light.cfg
    ### don't copy or process files
    y.pl --config=light --testonly

    ### only process files with changed MD5 signatures
    ### print file names while processing
    y.pl --update --verbose

    ### print debug information
    y.pl --debug


=head1 DESCRIPTION

=head1 BUGS

Update mode doesn't report changes in sidebars. When updating a file, its
parent should be updated as well to make sure navigation doesn't break.


=head1 AUTHOR

Morten Wulff <wulff@ratatosk.net>

Copyright (C) 2002, 2003 Morten Wulff

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.


=head1 VERSION

Version 2.2.0


=head1 SEE ALSO

L<HTML::Template>
